#include "streamline.h"
#include <math.h>

int _clamp(int vmin, int vmax, int v)
{
	return min(max(v,vmin), vmax);
}
bool isInImage(float px, float py, int w, int h)
{
	return (px>=0.0) && (px<=float(w-1)) && (py>=0.0) && (py<=float(h-1));
}
const float* getDirection(int w, int h, int cn, const float*flow,  int px, int py)
{
	const float *f = flow + _clamp(0,h-1,/*h-1-*/py)*w*cn + _clamp(0,w-1,px)*cn;
	return f;
}
float getConfidence(int w, int h, int cn, const float *confidence, int th, int px, int py)
{
	const float *f = confidence + _clamp(0,h-1,/*h-1-*/py)*w*cn + _clamp(0,w-1,px)*cn;
	return *(f+th);
}
void rotateDirection(float &dx, float &dy, float angle)
{
	float px = dx, py=dy;
	dx = cos(angle)*px - sin(angle)*py;
	dy = sin(angle)*px + cos(angle)*py;
}
////////////////////////////////////////////////////////////////////////// streamline stopping at low confidence (high curvature)
vector<Pot3f> streamline(int px, int py, int w, int h, int cn, float *flow, 
						  float *confidence, int th, float thre, float angle, int maxLen)
{
	vector<Pot3f> pos1, pos2;
	if(px<0 || py<0 || px>=w || py>=h)
	{
		cout << "seed should not out of the image." << endl;
		return pos1;
	}

	int cnt = 0;
	float seedx = float(px);
	float seedy = float(py);
	const float *fcenter = getDirection(w,h,cn,flow,px,py);
	float confcenter = getConfidence(w, h, cn, confidence, th, px, py);
	float conf = confcenter;
	float prevTangx = fcenter[1];
	float prevTangy = -fcenter[0];
	rotateDirection(prevTangx, prevTangy, angle);

	pos1.push_back(Pot3f(float(px),float(/*h-1-*/py),atan2(-prevTangy,prevTangx)));
	while(conf <= thre && cnt <maxLen && isInImage(seedx,seedy,w,h))
	{
		const float *fseed = getDirection(w,h,cn,flow,int(seedx+0.5),int(seedy+0.5));
		float tangx = fseed[1];
		float tangy = -fseed[0];
		rotateDirection(tangx, tangy, angle);
		float tangsign = tangx*prevTangx + tangy*prevTangy;
		if(tangsign<0.0)
		{
			tangx = -tangx;
			tangy= -tangy;
		}		
		seedx += tangx; //tangx>0.5?1: (tangx<-0.5?-1:0);
		seedy += tangy; //tangy>0.5?1: (tangy<-0.5?-1:0);

		pos1.push_back(Pot3f(seedx,/*float(h-1)-*/seedy,atan2(-tangy,tangx)));

		prevTangx = tangx;
		prevTangy = tangy;
		cnt ++;
		conf = getConfidence(w, h, cn, confidence, th, seedx, seedy);
	}

	cnt = 0;
	seedx = float(px);
	seedy = float(py);
	prevTangx = -fcenter[1];
	prevTangy = fcenter[0];
	rotateDirection(prevTangx, prevTangy, angle);
	conf = confcenter;	

	pos2.push_back(Pot3f(float(px),float(/*h-1-*/py),atan2(prevTangy,-prevTangx)));
	while(conf <= thre && cnt <maxLen && isInImage(seedx,seedy,w,h))
	{
		const float *fseed = getDirection(w,h,cn,flow,int(seedx+0.5),int(seedy+0.5));
		float tangx = fseed[1];
		float tangy = -fseed[0];
		rotateDirection(tangx, tangy, angle);
		float tangsign = tangx*prevTangx + tangy*prevTangy;
		if(tangsign<0.0)
		{
			tangx = -tangx;
			tangy= -tangy;
		}		
		seedx += tangx; //tangx>0.5?1: (tangx<-0.5?-1:0);
		seedy += tangy; //tangy>0.5?1: (tangy<-0.5?-1:0);

		pos2.push_back(Pot3f(seedx,/*float(h-1)-*/seedy,atan2(tangy,-tangx)));

		prevTangx = tangx;
		prevTangy = tangy;
		cnt ++;
		conf = getConfidence(w, h, cn, confidence, th, seedx, seedy);
	}

	pos1.insert(pos1.end(),pos2.begin(),pos2.end());
	pos1.push_back(Pot3f(float(pos1.size()-pos2.size()),float(pos2.size()),0.0));

//	cout << "left::" << pos1.size()-pos2.size()-1 << " right::" << pos2.size() << endl;
	return pos1;
}
vector<Pot3f> streamline(Pot2 pos, Pot2 wsz, int cn, float *flow, 
						  float *confidence, int th, float thre, float angle, int maxLen)
{
	return streamline(pos.x, pos.y, wsz.x, wsz.y, cn, flow, confidence, th, thre, angle, maxLen);
}
////////////////////////////////////////////////////////////////////////// hatching
vector<Pot3f> hatching(int hatLeftLen, int hatRightLen, float hatAngle, 
						int px, int py, int w, int h, int cn, float *flow)
{
	vector<Pot3f> res1, res2;
	const float *fcenter = getDirection(w,h,cn,flow,px,py);

	int cnt = 0;
	float seedx = float(px);
	float seedy = float(py);
	float prevGradx = fcenter[0];
	float prevGrady = -fcenter[1];
	rotateDirection(prevGradx,prevGrady,hatAngle);
	res1.push_back(Pot3f(float(px),float(/*h-1-*/py),atan2(-prevGrady,prevGradx)));
	while(cnt <hatLeftLen && isInImage(seedx,seedy,w,h))
	{
		const float *fseed = getDirection(w,h,cn,flow,int(seedx+0.5),int(seedy+0.5));
		float gradx = fseed[0];
		float grady = -fseed[1];
		rotateDirection(gradx,grady,hatAngle);
		float gradsign = gradx*prevGradx + grady*prevGrady;
		if(gradsign<0.0)
		{
			gradx = -gradx;
			grady = -grady;
		}
		seedx += gradx;
		seedy += grady;
		res1.push_back(Pot3f(seedx,/*float(h-1)-*/seedy,atan2(-grady,gradx)));

		prevGradx = gradx;
		prevGrady = grady;
		cnt++;
	}

	cnt = 0;
	seedx = float(px);
	seedy = float(py);
	prevGradx = -fcenter[0];
	prevGrady = fcenter[1];
	rotateDirection(prevGradx, prevGrady, hatAngle);
	while(cnt <hatRightLen && isInImage(seedx,seedy,w,h))
	{
		const float *fseed = getDirection(w,h,cn,flow,int(seedx+0.5),int(seedy+0.5));
		float gradx = fseed[0];
		float grady = -fseed[1];
		rotateDirection(gradx,grady,hatAngle);
		float gradsign = gradx*prevGradx + grady*prevGrady;
		if(gradsign<0.0)
		{
			gradx = -gradx;
			grady = -grady;
		}
		seedx += gradx;
		seedy += grady;
		res2.push_back(Pot3f(seedx,/*float(h-1)-*/seedy,atan2(-grady,gradx)));

		prevGradx = gradx;
		prevGrady = grady;
		cnt++;
	}

	vector<Pot3f> res(res1.size());
	for(int i=0; i<res1.size(); i++)
		res[i] = res1[res1.size()-1-i];

	res.insert(res.end(),res2.begin(),res2.end());
	return res;
}
vector<Pot3f> streamlineHatching(int hatStep, int hatLeftLen, int hatRightLen, float hatAngle, 
								  int px, int py, int w, int h, int cn, float *flow, 
								  float *confidence, int th, float thre, float angle, int maxLen)
{
	vector<Pot3f> pos1, pos2;
	if(px<0 || py<0 || px>=w || py>=h)
	{
		cout << "seed should not out of the image." << endl;
		return pos1;
	}

	int cnt = 0;
	float seedx = float(px);
	float seedy = float(py);
	const float *fcenter = getDirection(w,h,cn,flow,px,py);
	vector<Pot3f> onehatch = hatching(hatLeftLen,hatRightLen,hatAngle,seedx,seedy,w,h,cn,flow);
	pos1.insert(pos1.end(),onehatch.begin(),onehatch.end());

	float confcenter = getConfidence(w, h, cn, confidence, th, px, py);
	float conf = confcenter;
	float prevTangx = -fcenter[1];
	float prevTangy = -fcenter[0];
	rotateDirection(prevTangx, prevTangy, angle);
	while(conf <= thre && cnt <maxLen && isInImage(seedx,seedy,w,h))
	{
		const float *fseed = getDirection(w,h,cn,flow,int(seedx+0.5),int(seedy+0.5));
		float tangx = -fseed[1];
		float tangy = -fseed[0];
		rotateDirection(tangx, tangy, angle);
		float tangsign = tangx*prevTangx + tangy*prevTangy;
		if(tangsign<0.0)
		{
			tangx = -tangx;
			tangy= -tangy;
		}		
		seedx += tangx; //tangx>0.5?1: (tangx<-0.5?-1:0);
		seedy += tangy; //tangy>0.5?1: (tangy<-0.5?-1:0);

		if(cnt%hatStep==0 && cnt!=0)
		{
			onehatch = hatching(hatLeftLen,hatRightLen,hatAngle,seedx,seedy,w,h,cn,flow);
			pos1.insert(pos1.end(),onehatch.begin(),onehatch.end());
		}

		prevTangx = tangx;
		prevTangy = tangy;
		cnt ++;
		conf = getConfidence(w, h, cn, confidence, th, seedx, seedy);
	}

	cnt = 0;
	seedx = float(px);
	seedy = float(py);
	prevTangx = fcenter[1];
	prevTangy = fcenter[0];
	rotateDirection(prevTangx, prevTangy, angle);
	conf = confcenter;
	while(conf <= thre && cnt <maxLen && isInImage(seedx,seedy,w,h))
	{
		const float *fseed = getDirection(w,h,cn,flow,int(seedx+0.5),int(seedy+0.5));
		float tangx = -fseed[1];
		float tangy = -fseed[0];
		rotateDirection(tangx, tangy, angle);
		float tangsign = tangx*prevTangx + tangy*prevTangy;
		if(tangsign<0.0)
		{
			tangx = -tangx;
			tangy= -tangy;
		}		
		seedx += tangx; //tangx>0.5?1: (tangx<-0.5?-1:0);
		seedy += tangy; //tangy>0.5?1: (tangy<-0.5?-1:0);

		if(cnt%hatStep==0 && cnt!=0)
		{
			onehatch = hatching(hatLeftLen,hatRightLen,hatAngle,seedx,seedy,w,h,cn,flow);
			pos2.insert(pos2.end(),onehatch.begin(),onehatch.end());
		}

		prevTangx = tangx;
		prevTangy = tangy;
		cnt ++;
		conf = getConfidence(w, h, cn, confidence, th, seedx, seedy);
	}

	vector<Pot3f> pos(pos1.size());
	for(int i=0; i<pos1.size(); i++)
		pos[i] = pos1[pos1.size()-1-i];

	pos.insert(pos.end(),pos2.begin(),pos2.end());
	return pos;
}
vector<Pot3f> streamlineHatching(int hatStep, Pot2 hatLen, float hatAngle, 
								  Pot2 pos, Pot2 wsz, int cn, float *flow, 
								  float *confidence, int th, float thre, float angle, int maxLen)
{
	return streamlineHatching(hatStep, hatLen.x, hatLen.y, hatAngle, 
		pos.x, pos.y, wsz.x, wsz.y, cn, flow, confidence, th, thre, angle, maxLen);
}