#pragma once

#include <glew.h>
#include <QtGui/QtGui>
#include <QGLWidget>
#include <gl.h>
#include <glu.h>

#include <framebufferObject.h>
#include <texture2D.h>
#include <gpuProgram.h>
#include <gpuShader.h>

//#include "vectorization.h"
#include "tracking.h"
#include "skeleton.h"

using namespace gpu;

class ImagePaint : public QGLWidget {
	Q_OBJECT
		Q_PROPERTY(QImage image READ image WRITE setImage NOTIFY imageChanged(const QImage&)) 
		Q_PROPERTY(double zoom READ zoom WRITE setZoom NOTIFY zoomChanged(double)) 
		Q_PROPERTY(double resolution READ resolution WRITE setResolution NOTIFY resolutionChanged)
		Q_PROPERTY(double originX READ originX WRITE setOriginX NOTIFY originXChanged)
		Q_PROPERTY(double originY READ originY WRITE setOriginY NOTIFY originYChanged)
public:
	ImagePaint(QWidget *parent);
	~ImagePaint();

	virtual void restoreSettings(QSettings& settings);
	virtual void saveSettings(QSettings& settings);

	QSize imageSize() const { return image().size(); }
	const QImage& image() const { return m_image; }
	QPointF origin() const { return m_moveOrigin; }
	double originX() const { return m_moveOrigin.x(); }
	double originY() const { return m_moveOrigin.y(); }
	double zoom() const { return m_zoom; }
	double resolution() const { return m_resolution; }
	double scale() const { return m_zoom * m_resolution; }

	virtual QPointF view2image(const QPointF& p) const;
	virtual QPointF image2zoom(const QPointF& p) const;
	virtual QPointF image2view(const QPointF& p) const;
	virtual QTransform viewTransform(const QSizeF& sz, double zoom, const QPointF& origin) const;
	virtual float pt2px(float pt) const;

	void setOverlay(QWidget *overlay);

	public slots:
		void setImage(const QImage& image);
		void setOriginX(double value);
		void setOriginY(double value);
		void setZoom(double value);
		void setResolution(double value);
		void zoomIn();
		void zoomOut();
		void reset();
		void hold();
		void toggle();
		void copy();
		void savePNG(const QString& text=QString());

		void reloadShaders();
		void onIndexChanged(int);
		void refresh() { update(); }

		void nextFrame();
		void ComputeSkeletonChicked(bool isChecked);
		void setStreamline(bool);

signals:
		void imageChanged(const QImage&);
		void originXChanged(double);
		void originYChanged(double);
		void zoomChanged(double);
		void resolutionChanged(double);

public:
	void draw(QPainter& p, const QRectF& R, const QImage& image);

protected:
	virtual bool eventFilter( QObject *watched, QEvent *e );
	virtual void leaveEvent( QEvent *e );
	virtual void keyPressEvent( QKeyEvent *e );
	virtual void keyReleaseEvent( QKeyEvent *e );
	virtual void mousePressEvent( QMouseEvent *e );
	virtual void mouseMoveEvent( QMouseEvent *e );
	virtual void mouseReleaseEvent( QMouseEvent *e );
	virtual void wheelEvent(QWheelEvent *e);

	virtual void initializeGL();
	virtual void resizeGL(int w, int h);
	virtual void paintGL();

	inline void drawQuad(int width, int height, int xoff=0, int yoff=0);
	inline void drawQuadScreen(int width, int height, int xoff=0, int yoff=0);
	inline void swapToLocalMode(int w, int h);
	inline void swapToWorldMode();

protected:
	/// move tool
	QPointF m_moveOrigin;
	QPointF m_moveStart;
	QPointF m_moveOriginKeep;
	double m_zoom;

	/// zoom tool
	QPointF m_zoomOrigin;
	QPointF m_zoomOriginKeep;
	double m_resolution;	

	enum Mode { MODE_MOVE=0, MODE_ROTO };
	Mode m_mode;

	bool m_spacePressed;
	/// interaction
	QCursor m_cursor;
	Qt::MouseButton m_dragButton;	
	QPoint	m_currPosWnd;
	QPointF  m_currPosImg;

	/// image buffer
	int m_index;
	QImage m_image;
	bool	m_imgLoaded;
	QWidget *m_overlay;

	/// size
	QSize	m_wndSize;
	QSize	m_imgSize;
	int		m_numSpot;

	/// for dynamic focus+context
	bool	m_keyCtrl;
	bool	m_bZoomInOut;
	QTimer	*_timer;
	int		_millisec;

	int		m_frameCurr;
	int		m_frameCount;

	skeleton::SkeletonGraph m_sktGraph;
	bool	m_bSkeletonDisplay;

	////////////////////////////////////////// GPU
private:
	enum {FBO1=0};

	enum {T_ORG=0,	T_GRADIENT, T_GRADIENT_S, T_PROFILE, T_PROJECT, T_PEARL, T_FIN};

	enum {P_GRADIENT, P_GRADIENT_S, P_PROFILE,  P_FIN};

	static const int NB_FBO = 1;
	static const int NB_TEX = 7;
	static const int NB_PROG = 4;

	FramebufferObject	  *_fbo[NB_FBO];
	GPUProgram            *_prog[NB_PROG];
	FloatTexture2D        *_tex[NB_TEX];

	/////////////////////////////////// parameters
	//int			m_smoothSizeUsing;
	int			m_gradSize;
	int			m_gradSmoothSize;
	double      m_widthf;
	int			m_trackingLength;

	int			m_zoomType;
	bool		m_streamlineDrawing;

public:
	int* gradSizePtr() {return &m_gradSize;}
	int* gradSmoothSizePtr() {return &m_gradSmoothSize;}
	double* widthfPtr() {return &m_widthf;}
	int* trackingLengthPtr() {return &m_trackingLength;}

	int* zoomTypePtr() {return &m_zoomType;}

private:
	void initGPU();
	void finalizeGPU();

	void initFBO();
	void cleanFBO();

	void initShaders();
	void setShaderParameters();
	void cleanShaders();

	void setTextureDefault(int tex_id, float val);
	//void initGaussWeightsTex(int tex_id, int smsz, float supportFactor);
	void loadTexture(int tex_id, const QImage &qimg);
	void initNoiseTex(int tex_id, const QSize &imgSize);
	//bool loadSegmentation(int tex_id, const char*filename);
	//void loadImportanceMap(int tex_id, const char*filename);

	void drawStreamline();
	void drawStreamlineOne(const vector<tracking::Pot3f> &pnt, 
		const tracking::Pot3f &clr, const tracking::Pot3f &clrStart);
	void updateProfileLabel(const float *pGrad, const float *pProf, const float *pProj,
		int proWidth, int px, int py, float dLeft, float dRight);

	void computeSkeletonGraph(skeleton::SkeletonGraph &sktGraph);
	skeleton::Edge tracking(float px, float py, float tangx, float tangy, int w, int h, int cn, float *grad, float *flow, vector<skeleton::Vertex> &vertexRecord, vector<QPointF> &angleRecord);
	void ImagePaint::trackingEdge(skeleton::Edge &pos, float px, float py,float &tangx, float &tangy, int w, int h, int cn, float *flow,int maxLen);
	const float* ImagePaint::centerDirection(int w, int h, int cn, float px, float py, float* grad, float *flow);
};

inline void ImagePaint::drawQuad(int width, int height, int xoff/* =0 */, int yoff/* =0 */)
{
	glBegin(GL_QUADS);
	glTexCoord2i(0,0);  glVertex2i(xoff, yoff);
	glTexCoord2i(1,0);  glVertex2i(width+xoff, yoff);
	glTexCoord2i(1,1);  glVertex2i(width+xoff, height+yoff);
	glTexCoord2i(0,1);  glVertex2i(xoff, height+yoff);
	glEnd();
}

inline void ImagePaint::drawQuadScreen(int width, int height, int xoff/* =0 */, int yoff/* =0 */)
{
	glBegin(GL_QUADS);
	glTexCoord2i(0,1);  glVertex2i(xoff, yoff);
	glTexCoord2i(1,1);  glVertex2i(width+xoff, yoff);
	glTexCoord2i(1,0);  glVertex2i(width+xoff, height+yoff);
	glTexCoord2i(0,0);  glVertex2i(xoff, height+yoff);
	glEnd();
}

inline void ImagePaint::swapToLocalMode(int w, int h)
{
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0,w,0,h,-1,1);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glViewport(0,0,w,h);
	glLoadIdentity();
}

inline void ImagePaint::swapToWorldMode()
{
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}
